/**
 *    @file  libhttp.c
 *   @brief  HTTP parsing and building library
 *
 *  @author  François Cerbelle (Fanfan), francois@cerbelle.net
 *
 *  @internal
 *       Created:  15/11/2024
 *      Revision:  none
 * Last modified:  2024-11-20 20:59
 *      Compiler:  gcc
 *  Organization:  Cerbelle.net
 *     Copyright:  Copyright (c) 2024, François Cerbelle
 *
 *  This source code is released for free distribution under the terms of the
 *  GNU General Public License as published by the Free Software Foundation.
 */

#include "libhttp.h"
#include "sclist.h"
#include "base64.h"                             /* Base64 encoder and decoder */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <assert.h>

typedef struct HTTP_s {
    sclist_t* headers;
    char* body;
} HTTP_t;

typedef struct HTTPHeader_s {
    char* name;
    char* value;
} HTTPHeader_t;

HTTPHeader_t* HTTPHeader_new(const char* name, const char* value) {
    HTTPHeader_t* header;

    assert(name);
    assert(value);

    if (NULL==(header=malloc(sizeof(struct HTTPHeader_s)))) {
        perror("HTTPHeader_new");
    } else {
        header->name = strdup(name);
        header->value = strdup(value);
    }
    return header;
}

void          HTTPHeader_del(HTTPHeader_t* header) {
    assert(header);
    assert(header->name);
    assert(header->value);

    free(header->name);
    free(header->value);
    free(header);
}

HTTPHeader_t* HTTPHeader_basicauth(const char* login, const char* pass) {
    assert(login);
    assert(pass);

    /* Prepare basic authentication */
    char* auth_encoded;
    char* auth;
    {
        if (NULL==(auth=malloc(strlen("Basic ")+strlen(login)+1+strlen(pass)+1))) {
            perror("HTTPHeader_basicauth");
            return NULL;
        };
        /* strcpy/strcat expected to be faster than sprintf */
        strcpy(auth,login);
        strcat(auth,":");
        strcat(auth,pass);
        auth_encoded=base64_encode(auth);
        free(auth);
    }

    if (NULL==(auth=malloc(strlen("Basic ")+strlen(auth_encoded)+1))) {
        perror("HTTPHeader_basicauth");
        free(auth_encoded);
        return NULL;
    };
    strcpy(auth,"Basic ");
    strcat(auth, auth_encoded);
    return HTTPHeader_new("Authorization",auth);
}

char*         HTTPHeader_getname(HTTPHeader_t* header) {
    assert(header);
    return header->name;
}

char*         HTTPHeader_getvalue(HTTPHeader_t* header) {
    assert(header);
    return header->value;
}

HTTP_t* HTTP_new() {
    HTTP_t* retval;

    if (NULL==(retval=malloc(sizeof(struct HTTP_s)))) {
        perror("HTTP_new HTTP_s");
        return NULL;
    }

    if (NULL==(retval->body=malloc(1))) {
        perror("HTTP_new body");
        free(retval);
        return NULL;
    };
    retval->body[0]=0;

    if (NULL==(retval->headers = sclist_new())) {
        perror("HTTP_new headers");
        free(retval->body);
        free(retval);
        return NULL;
    }

    return retval;
}








typedef struct HTTPHeaders_s {
    char* name;
    char* value;
} HTTPHeaders_t;

void HTTP_del(HTTP_t* http) {
    sclistrecord_t* current;

    assert(http);
    assert(http->headers);
    assert(http->body);

    /* Use sclist_remrecord() and sclist_del() */
    current = sclist_firstrecord(http->headers);
    while (current) {
        HTTPHeaders_t* header=sclist_getvalue(current);
        if (header->name)
            free(header->name);
        if (header->value)
            free(header->value);
        current = sclist_nextrecord(current);
    }

    free(http->body);
    free(http);
}


void HTTP_addheader(HTTP_t* http, const char* name, const char* value) {
    HTTPHeaders_t* headers;

    assert(http);
    assert(http->headers);
    assert(http->body);

    if (NULL==(headers=malloc(sizeof(struct HTTPHeaders_s)))) {
        perror("HTTP_addheader");
        exit(EXIT_FAILURE);
    }
    if (NULL==(headers->name=malloc(strlen(name)+1))) {
        perror("HTTP_addheader name malloc");
        exit(EXIT_FAILURE);
    }
    strcpy(headers->name,name);
    if (NULL==(headers->value=malloc(strlen(value)+1))) {
        perror("HTTP_addheader value malloc");
        exit(EXIT_FAILURE);
    }
    strcpy(headers->value,value);
    sclist_addrecord(http->headers,headers);
}

char* HTTP_getheaders(const HTTP_t* http) {
    sclistrecord_t* current = NULL;
    char* retval;

    assert(http);
    assert(http->headers);
    assert(http->body);

    if (NULL==(retval=malloc(1))) {
        perror("HTTP_getheaders initial malloc");
        exit(EXIT_FAILURE);
    };
    retval[0]=0;

    current = sclist_firstrecord(http->headers);
    while (current) {
        HTTPHeaders_t* header=sclist_getvalue(current);
        if (header->name && header->value) {
            char* newretval;
            if (NULL==(newretval=realloc(retval,strlen(retval)+strlen(header->name)+2+strlen(header->value)+2+1))) {
                perror("HTTP_getheaders inner loop");
                exit(EXIT_FAILURE);
            }
            retval=newretval;
            strcat(retval,header->name);
            strcat(retval,": ");
            strcat(retval,header->value);
            strcat(retval,"\r\n");
        }
        current=sclist_nextrecord(current);
    }
    return retval;
}

char* HTTP_getrequest(const HTTPMethod_t method, const char* uri, const HTTPVersion_t version) {
    char* retval;

    assert(method<=HTTPMETHOD_INVALID);
    assert(version<=HTTPVERSION_INVALID);

    if (NULL==(retval=malloc(7+1+strlen(uri)+1+8+1))) {
        perror("HTTP_getrequest");
        exit(EXIT_FAILURE);
    }
    retval[0]=0;

    switch(method) {
    case HTTPMETHOD_GET:
        strcat(retval, "GET");
        break;;
    case HTTPMETHOD_HEAD:
    case HTTPMETHOD_POST:
    case HTTPMETHOD_PUT:
    case HTTPMETHOD_DELETE:
    case HTTPMETHOD_CONNECT:
    case HTTPMETHOD_OPTIONS:
    case HTTPMETHOD_TRACE:
    case HTTPMETHOD_PATCH:
    case HTTPMETHOD_INVALID:
    default:
        fprintf(stderr,"HTTP method not supported\n");
        exit(EXIT_FAILURE);
        break;;
    }
    strcat(retval," ");

    /* NULL URI allowed, as an empty string */
    if (uri)
        strcat(retval, uri);

    strcat(retval," HTTP/");
    switch(version) {
    case HTTPVERSION_HTTP11:
    case HTTPVERSION_HTTP11b:
        strcat(retval, "1.1");
        break;;
    case HTTPVERSION_HTTP09:
    case HTTPVERSION_HTTP10:
    case HTTPVERSION_HTTP2:
    case HTTPVERSION_HTTP3:
    case HTTPVERSION_INVALID:
    default:
        fprintf(stderr,"HTTP version not supported\n");
        exit(EXIT_FAILURE);
        break;;
    }
    strcat(retval, "\r\n");
    return retval;
}

void HTTP_setbody(HTTP_t* http, const char* body) {
    char* newbody;

    assert(http);
    assert(http->body);

    if (NULL==(newbody=realloc(http->body,strlen(body)+1))) {
        perror("HTTP_setbody");
        exit(EXIT_FAILURE);
    }

    http->body = newbody;
    strcpy(http->body, body);
}

char* HTTP_getbody(HTTP_t* http) {
    assert(http);
    assert(http->body);

    return http->body;
}
